#
#
# Name:   stabc.py - stability checking for communicating automata.
#
#
################################################################################

# All temporary files are copied to this directory 
# All computations are achieved in this directory as well

mytmp="tmp/"

##
# A label is a couple (message, direction)
class Label:
    def __init__(self,message,direction):

        self._message=message.lower()
	self._direction=direction

    def getMessage(self):
        return self._message

    def getDirection(self):
        return self._direction

    def opposite(self):
        if (self._direction=="EM"):
            return "REC"
        else:
            return "EM"

##
# A state is an identifier
class State:

    def __init__(self,name):
        self._name=name

    def getName(self):
        return self._name

##
# A transition is a tuple (source state, label, target state)
class Transition:
    def __init__(self,src,label,tgt):
        self._src=src
        self._tgt=tgt
        self._label=label

    def getSource(self):
        return self._src

    def getTarget(self):
        return self._tgt

    def getLabel(self):
        return self._label

##
# A peer LTS is a tuple (name, init, S, T)
class Peer:
    def __init__(self,name,init):
        self._name = name
        self._s0 = init
        self._S = []
        self._T = []

    def addState(self, state):
        if not state.getName() in map(lambda s: s.getName(), self._S):
            self._S.append(state)

    def addTransition(self,transition):
        self._T.append(transition)

    def getName(self):
        return self._name

    # Builds alphabet
    def buildAlphabet(self):
        alpha=[]
        for t in self._T:
            if (t.getLabel().getMessage()!="i"):
                res=t.getLabel().getMessage()+"_"+t.getLabel().getDirection()
                if not(res in alpha):
                    alpha.append(res)
        return alpha

    # Builds alphabet without direction
    def buildAlphabetNoDirection(self):
        alpha=[]
        for t in self._T:
            res=t.getLabel().getMessage()
            if (res!="i"):
                if not(res in alpha):
                    alpha.append(res)
        return alpha

    # Builds alphabet adding EM suffixes systematically
    def buildAlphabetEmissionOnly(self):
        alpha=[]
        for t in self._T:
            if (t.getLabel().getMessage()!="i"):
                res=t.getLabel().getMessage()+"_EM"
                if not(res in alpha):
                    alpha.append(res)
        return alpha

    # Builds alphabet for the corresponding queue
    def buildAlphabetQueue(self):
        alpha=[]
        for t in self._T:
            if (t.getLabel().getDirection()=="REC"):
                msg=t.getLabel().getMessage()
                if not(msg in alpha):
                    alpha.append(msg)
        return alpha

    # Dumps alphabet for queues
    def dumpAlphabetQueue(self,alpha,f,any):
        l=0
        alpha = sorted (alpha)
        max=len(alpha)
        if alpha!=[]:
            f.write(" [")
            for e in alpha:
                f.write(e+"_EM")
                if any:
                    f.write(":any")
                f.write(",")
                f.write(e+"_REC")
                if any:
                    f.write(":any")
                l=l+1
                if l<max:
                    f.write(",")
            f.write("] ")

    # Dumps alphabet
    def dumpAlphabet(self,alpha,f,any):
        l=0
        alpha = sorted (alpha)
        max=len(alpha)
        if alpha!=[]:
            f.write(" [")
            for e in alpha:
                f.write(e)
                if any:
                    f.write(":any")
                l=l+1
                if l<max:
                    f.write(",")
            f.write("] ")

    # Dumps alphabet with REC
    def dumpAlphabetWithREC(self,alpha,f,any):
        l=0
        alpha = sorted (alpha)
        max=len(alpha)
        if alpha!=[]:
            for e in alpha:
                f.write(e+"_REC")
                if any:
                    f.write(":any")
                l=l+1
                if l<max:
                    f.write(",")

    # Returns the list of transitions with sid as source
    def getTransitions(self,sid):
        subT=[]
        for t in self._T:
            if (sid==t.getSource().getName()):
               subT.append(t)
        return subT

    # Generates an LNT process for a state in the peer LTS
    def state2lnt(self,sid,alpha,f):
        f.write("process "+self._name+"_"+sid)
        self.dumpAlphabet(alpha,f,True)
        f.write(" is \n")
        subT=self.getTransitions(sid)
        max=len(subT)
        if (max>0):
          f.write(" select \n")
          i=0
          for t in subT:
              if (t.getLabel().getMessage()!="i"):
                  res=t.getLabel().getMessage()+"_"+t.getLabel().getDirection()
              else:
                  res=t.getLabel().getMessage()
              f.write("  "+res+"; ")
              f.write(self._name+"_"+t.getTarget().getName())
              self.dumpAlphabet(alpha,f,False)
              f.write("\n")
              i=i+1
              if i<max:
                 f.write("\n[]\n")
          f.write(" end select \n")
        else:
          f.write(" null \n")
        f.write("end process \n\n")

    # returns the length of the longest sequence of emissions (LLSE)
    # returns -1 if infinite
    def longestEmissionPath(self):
        # optim: we could be more efficient to compute this result..
        res=0
        for s in self._S:
            raux=self.LEPaux(s,[],0,0)
            if (raux>res):
                res=raux
        return res

    # checks if there is a cycle only with emissions
    def existsEmissionCycle(self, state, visited, visitedincycle):
        if (state.getName() in visitedincycle): # a cycle exists
            #print state.getName()
            #print visitedincycle
            return True
        if (state.getName() in visited): # already visited
            return False
        else:
            subT=self.getTransitions(state.getName())
            if (subT!=[]):
                visited.append(state.getName())
                for t in subT:
                    if (t.getLabel().getDirection()=="EM"):
                        visitedincycle.append(state.getName())
                        res=self.existsEmissionCycle(t.getTarget(),visited,visitedincycle)
                        visitedincycle.pop()
                    else:
                        res=self.existsEmissionCycle(t.getTarget(),visited,[])
                    if res:
                        return res
            else: # no successor, so not a cycle
                return False

    # auxiliary method for computing LLSE
    # length is the current longest, max is the longest so far
    def LEPaux(self, state, visited, length, maxi):
        if (state.getName() in visited): # already visited
            return maxi
        else:
            subT=self.getTransitions(state.getName())
            if (subT!=[]):
                visited.append(state.getName())
                for t in subT:
                    #maxloop=max
                    if (t.getLabel().getDirection()=="EM"):
                        if ((length+1)>maxi):
                            maxitmp=length+1
                        else:
                            maxitmp=maxi
                        res=self.LEPaux(t.getTarget(),visited,length+1,maxitmp)
                    else:
                        res=self.LEPaux(t.getTarget(),visited,0,maxi)
                    #if (res>maxloop):
                    #    maxloop=res
                return res
            else: # no successor
                return maxi

    # returns the list of all sequences of emissions in this peer
    def computeEmissionSequences(self):
        # optim: we could be more efficient to compute this result..
        res=[]
        for s in self._S:
            lseq=self.CESaux(s,[],"",[])
            for seq in lseq:
                if not(seq in res):
                    res.append(seq) # we compute more sequences than necessary but does not harm the result
        return res

    # auxiliary method for computing the list of all emission sequences
    # current is the current sequence, lseq keeps the list of sequences
    def CESaux(self, state, visited, current, lseq):
        if (state.getName() in visited): # already visited
            if (current!="") and not(current in lseq):
                lseq.append(current)
            return lseq
        else:
            subT=self.getTransitions(state.getName())
            if (subT!=[]):
                visited.append(state.getName())
                #res=[]
                for t in subT:
                    if (t.getLabel().getDirection()=="EM"):
                        currenttmp=current
                        if (current==""):
                            current=t.getLabel().getMessage()
                            
                        else:
                            current=current+"."+t.getLabel().getMessage()
                        res=self.CESaux(t.getTarget(),visited,current,lseq)
                        current=currenttmp
                    else:
                        current=""
                        res=self.CESaux(t.getTarget(),visited,current,lseq)
                if (current!="") and not(current in lseq):
                    lseq.append(current)
                return lseq
            else: # no successor
                if (current!="") and not(current in lseq):
                    lseq.append(current)
                return lseq


##
# Asynchronous compositions
class Async:
    def  __init__(self, name):
        self._name=name
        self._peers=[]

    def aut2py4peers(self,nbpeers):
        import re
	from subprocess import call

	# generate python for all "peerI" from 1 to nbpeers
	ind = 1
        cdm="cp "+self._name+"/peer*.aut " + mytmp
	call(cdm, shell=True)
	while (ind<=nbpeers):
                
		nfile=mytmp+"peer"+str(ind)+".aut"
		nfileaux=mytmp+"peer"+str(ind)+"_aux.aut"

                # removes unnecessary characters
                string=open(nfile).read()
                string=string[4:]
                new_str=re.sub('[^a-zA-Z0-9\n\.\ \_]', '', string)
                open(nfileaux,'w').write(new_str)

                # reads all the file contents and storing into data
		lines=tuple(open(nfileaux, "r"))
		data=[[0 for x in xrange(3)] for x in xrange(len(lines))]
		for j in range(0, len(lines)):
			data[j]=[s.strip() for s in lines[j].split(' ')]
		# print data

                pname="peer"+str(ind)

                # preprocessing on states
                setStates=list()
                setStates.append(data[0][0])
                for i in range(1,len(data)):
                    if data[i][0] not in setStates:
                        setStates.append(data[i][0])
                    if data[i][2] not in setStates:
                        setStates.append(data[i][2])

                # builds the peer objects
                peeraux=Peer(pname,State("s"+str(setStates[0])))
                for s in setStates:
                    peeraux.addState(State("s"+str(s)))
                for i in range(1, len(data)):
                    if data[i][1]!='exit':
                        tem1=data[i][1].split('_')
                        if (tem1[0]=="i"):
                            tem1.append("") # dummy direction
                        peeraux.addTransition(Transition(State("s"+str(data[i][0])),Label(tem1[0],tem1[1]),State("s"+str(data[i][2]))))

                self._peers.append(peeraux)
                ind=ind+1
		# f.close()

    # Dumps list of strings
    def dumpStringList(self,f,alpha):
        max=len(alpha)
        l=0
        for e in alpha:
            f.write(e)
            l=l+1
            if l<max:
               f.write(",")

    # Generates a few types (string, fifo) and functions
    def generateDatatypesFIFO(self,alpha,f):
        f.write("type Message is \n")
        self.dumpStringList(f,alpha)
        f.write("\n")
        f.write("with \"==\", \"!=\"\n")
        f.write("end type \n\n")
        f.write("type Queue is list of Message \n")
        f.write("with \"==\", \"!=\"\n")
        f.write("end type \n\n")
        f.write("type BoundedQueue is bqueue (queue: Queue, bound: Nat) \n")
        f.write("with \"==\", \"!=\"\n")
        f.write("end type \n\n")
        f.write("function insert (m: Message, q: Queue): Queue is \n")
        f.write("         case q in \n")
        f.write("         var hd: Message, tl: Queue in \n")
        f.write("             nil         -> return cons(m,nil) \n")
        f.write("           | cons(hd,tl) -> return cons(hd,insert(m,tl)) \n")
        f.write("         end case \n")
        f.write("end function \n\n")
        f.write("function ishead (m: Message, q: Queue): Bool is \n")
        f.write("         case q in \n")
        f.write("         var hd: Message, tl: Queue in \n")
        f.write("             nil         -> return false \n")
        f.write("           | cons(hd,tl) -> return (m==hd) \n")
        f.write("         end case \n")
        f.write("end function \n\n")
        f.write(" (* function remove suppresses the first element for FIFO buffers *) \n")
        f.write(" (* hence parameter m is not used *) \n")
        f.write("function remove (m: Message, q: Queue): Queue is \n")
        f.write("         case q in \n")
        f.write("         var hd: Message, tl: Queue in \n")
        f.write("             nil         -> return nil \n")
        f.write("           | cons(hd,tl) -> return tl \n")
        f.write("         end case \n")
        f.write("end function \n\n")
        f.write("function count (q: Queue): Nat is \n")
        f.write("         case q in \n")
        f.write("         var hd: Message, tl: Queue in \n")
        f.write("             nil         -> return 0 \n")
        f.write("           | cons(hd,tl) -> return (1+count(tl)) \n")
        f.write("         end case \n")
        f.write("end function \n\n")
        f.write("function bisfull (bq: BoundedQueue): Bool is \n")
        f.write("  return ((count(bq.queue))==bq.bound) \n")
        f.write("end function \n\n")
        f.write("function binsert (m: Message, bq: BoundedQueue): BoundedQueue is \n")
        f.write("  if bisfull(bq) then  \n")
        f.write("     return bq  \n")
        f.write("  else  \n")
        f.write("     return bqueue(insert(m,bq.queue),bq.bound)  \n")
        f.write("  end if \n")
        f.write("end function \n\n")
        f.write("function bishead (m: Message, bq: BoundedQueue): Bool is \n")
        f.write("  return ishead(m,bq.queue) \n")
        f.write("end function \n\n")
        f.write("function bremove (m: Message, bq: BoundedQueue): BoundedQueue is \n")
        f.write("  return bqueue(remove(m,bq.queue),bq.bound) \n")
        f.write("end function \n\n")
        f.write("function bcount (bq: BoundedQueue): Nat is \n")
        f.write("  return count(bq.queue) \n")
        f.write("end function \n\n")

    # Generates a few types (string, fifo) and functions
    def generateDatatypesBAG(self,alpha,f):
        f.write("type Message is \n")
        self.dumpStringList(f,alpha)
        f.write("\n")
        f.write("with \"==\", \"!=\"\n")
        f.write("end type \n\n")
        f.write("type Queue is list of Message \n")
        f.write("with \"==\", \"!=\"\n")
        f.write("end type \n\n")
        f.write("type BoundedQueue is bqueue (queue: Queue, bound: Nat) \n")
        f.write("with \"==\", \"!=\"\n")
        f.write("end type \n\n")
        f.write("function insert (m: Message, q: Queue): Queue is \n")
        f.write("         case q in \n")
        f.write("         var hd: Message, tl: Queue in \n")
        f.write("             nil         -> return cons(m,nil) \n")
        f.write("           | cons(hd,tl) -> return cons(hd,insert(m,tl)) \n")
        f.write("         end case \n")
        f.write("end function \n\n")
        f.write("(* In the case of BAG buffers, this function checks the presence of a msg m *) \n")
        f.write("(* in the queue q, not necessarily at the first place *) \n")
        f.write("function ishead (m: Message, q: Queue): Bool is \n")
        f.write("         case q in \n")
        f.write("         var hd: Message, tl: Queue in \n")
        f.write("             nil         -> return false \n")
        f.write("           | cons(hd,tl) -> return (m==hd) or ishead(m,tl) \n")
        f.write("         end case \n")
        f.write("end function \n\n")
        f.write("(* In the case of BAG buffers, this function removes one instance of the msg m *) \n")
        f.write("(* This instance can be stored anywhere in the queue *) \n")
        f.write("function remove (m: Message, q: Queue): Queue is \n")
        f.write("         case q in \n")
        f.write("         var hd: Message, tl: Queue in \n")
        f.write("             nil         -> return nil \n")
        f.write("           | cons(hd,tl) -> if (hd==m) then return tl else return cons(hd,remove(m,tl)) end if \n")
        f.write("         end case \n")
        f.write("end function \n\n")
        f.write("function count (q: Queue): Nat is \n")
        f.write("         case q in \n")
        f.write("         var hd: Message, tl: Queue in \n")
        f.write("             nil         -> return 0 \n")
        f.write("           | cons(hd,tl) -> return (1+count(tl)) \n")
        f.write("         end case \n")
        f.write("end function \n\n")
        f.write("function bisfull (bq: BoundedQueue): Bool is \n")
        f.write("  return ((count(bq.queue))==bq.bound) \n")
        f.write("end function \n\n")
        f.write("function binsert (m: Message, bq: BoundedQueue): BoundedQueue is \n")
        f.write("  if bisfull(bq) then  \n")
        f.write("     return bq  \n")
        f.write("  else  \n")
        f.write("     return bqueue(insert(m,bq.queue),bq.bound)  \n")
        f.write("  end if \n")
        f.write("end function \n\n")
        f.write("function bishead (m: Message, bq: BoundedQueue): Bool is \n")
        f.write("  return ishead(m,bq.queue) \n")
        f.write("end function \n\n")
        f.write("function bremove (m: Message, bq: BoundedQueue): BoundedQueue is \n")
        f.write("  return bqueue(remove(m, bq.queue),bq.bound) \n")
        f.write("end function \n\n")
        f.write("function bcount (bq: BoundedQueue): Nat is \n")
        f.write("  return count(bq.queue) \n")
        f.write("end function \n\n")

    # Dumps the alphabet for a peer queue process
    def generateAlphaQueue(self,alpha,f,any):
            max=len(alpha)
            i=0
            if alpha!=[]:
               f.write(" [")
               for m in alpha:
                   f.write(m)
                   if any:
                      f.write(":any")
                   f.write(",")
                   f.write(m+"_REC")
                   if any:
                      f.write(":any")
                   i=i+1
                   if i<max:
                      f.write(",")
               f.write("] ")

    # Generates an LNT process for a peer queue
    def generateQueue(self,p,f):

            f.write("process ")
            pname="queue_"+p.getName()
            f.write(pname)
            alpha=p.buildAlphabetQueue()
            p.dumpAlphabetQueue(alpha,f,True)
            f.write(" (bq: BoundedQueue) is \n ")

            f.write(" select \n")
            for m in alpha:
                f.write("    if not(bisfull(bq)) then ")
                f.write(m+"_EM ; "+pname)
                p.dumpAlphabetQueue(alpha,f,False)
                f.write(" (binsert("+m+"_REC,bq)) else stop end if\n")
                f.write("    [] \n")
                gaterec=m+"_REC"
                f.write("    if bishead("+m+"_REC,bq) then "+gaterec+" ; "+pname)
                p.dumpAlphabetQueue(alpha,f,False)
                f.write(" (bremove("+m+"_REC,bq)) else stop end if\n")
                f.write("    [] \n")

            f.write("    null \n ")
            f.write(" end select \n ")

            f.write("end process \n\n")

    # Computes the union of 2 lists
    def union(self,l1,l2):
        l=[]
        for e in l1:
            if not(e in l2):
               l.append(e)
        l.extend(l2)
        return l

    # Adds both directions to an alphabets
    def addBothDirections(self,alpha):
        res=[]
        for e in alpha:
            res.append(e+"_EM")
            res.append(e+"_REC")
        return res

    # Generates an LNT process for a peer queue
    def generatePeerQueue(self,p,f,bound):
        f.write("process ")
        pname="peer_queue_"+p.getName()+"_"+str(bound)
        f.write(pname)
        # builds alphabets
        alphapeer=p.buildAlphabet()
        alphaqueue=p.buildAlphabetQueue()
        alphaqueuebis=self.addBothDirections(alphaqueue)
        alpha=self.union(alphapeer,alphaqueuebis)
        p.dumpAlphabet(alpha,f,True)
        f.write(" is \n")
        f.write("  par ")
        synchro=p.buildAlphabetQueue()
        if synchro!=[]:
           p.dumpAlphabetWithREC(synchro,f,False)
           f.write(" in \n")
        f.write (p.getName())
        p.dumpAlphabet(alphapeer,f,False)
        f.write("\n")
        f.write("    || \n")
        f.write("    queue_"+p.getName())
        p.dumpAlphabetQueue(alphaqueue,f,False)
        f.write("(bqueue(nil,"+str(bound)+")) \n")
        f.write("  end par \n")
        f.write("end process \n\n")

    # Generates an LNT file given a set of peers
    def peers2lnt(self, nbpeers, nfile, bound, synchronizability=True, increment=1, cmodel="fifo"):
        filename=mytmp+nfile+".lnt"
        f=open(filename, 'w')
        f.write("module "+self._name+" with \"get\" is \n\n")
        f.write("(* LNT processes for peers *)\n\n")
        alphaem=[]
        for p in self._peers:
            alpha=p.buildAlphabet()
            f.write("process "+p.getName())
            p.dumpAlphabet(alpha,f,True)
            f.write(" is \n")
            f.write(p.getName()+"_"+p._s0.getName())
            p.dumpAlphabet(alpha,f,False)
            f.write("\n")
            f.write("end process \n\n")
            for k in p._S:
               p.state2lnt(k.getName(),alpha,f)
            # accumulates alphabets
            for e in alpha:
                if not(e in alphaem):
                    alphaem.append(e)

        # datatype generation
        if (cmodel=="fifo"):
            f.write("(* Datatypes for messages and FIFO buffers *)\n\n")
            self.generateDatatypesFIFO(alphaem,f)
        else:
            f.write("(* Datatypes for messages and BAG buffers *)\n\n")
            self.generateDatatypesBAG(alphaem,f)

        if synchronizability:
            for p in self._peers:
                self.generateQueue(p,f)
            for p in self._peers:
                self.generatePeerQueue(p,f,1)
        else:
            for p in self._peers:
                self.generateQueue(p,f)
            for p in self._peers:
                self.generatePeerQueue(p,f,bound)
            for p in self._peers:
                self.generatePeerQueue(p,f,bound+increment)

        f.write("end module \n")
        f.close()

    # checks the existence of a cycle of emissions in one peer
    def existsEmissionCyclePeers(self):
        res=0
        for p in self._peers:
            if p.existsEmissionCycle(p._s0,[],[]):
                return -1
        return res

    # returns the length of the longest sequence of emissions (LLSE) for all peers
    # returns -1 if infinite
    def longestEmissionPathPeers(self):
        res=0
        for p in self._peers:
            tmp=p.longestEmissionPath()
            if (tmp>res):
                res=tmp
        return res

    # returns the total number of emissions for all peers
    # returns -1 if infinite
    def numberOfEmissionsPeers(self):
        res=0
        for p in self._peers:
            if p.existsEmissionCycle(p._s0,[],[]):
                return -1
            else:
                for t in p._T:
                    if (t.getLabel().getDirection()=="EM"):
                        res=res+1
        return res


    # counts relevant messages in the list of sequences wrt an alphabet
    def countEmInAlpha(self,lseq,alpha):
        res=0
        for ch in lseq:
            seq2=ch.split('.')
            restmp=0
            for msg in seq2:
                if (msg in alpha):
                    restmp=restmp+1
            if (restmp>res):
                res=restmp
        return res

    # returns the total number of emissions in all peers destinated to p
    # to do so, analyze all sequences of emissions
    def checkAllEmissionsToPeer(self,peername,alpharec):
        res=0
        for p in self._peers:
            if (peername!=p.getName()):
                seqem=p.computeEmissionSequences()
                resaux=self.countEmInAlpha(seqem,alpharec)
                res=res+resaux
        return res

    # returns the total number of receptions in a specific peer
    def allReceptionsInPeer(self,peername):
        res=[]
        for p in self._peers:
            if (peername==p.getName()):
                for t in p._T:
                    if (t.getLabel().getDirection()=="REC"):
                        if not(t.getLabel().getMessage() in res):
                            res.append(t.getLabel().getMessage())
        return res

    # returns the max number of received emissions for all peers
    # returns -1 if infinite
    def numberOfReceivedEmissionsPeers(self):
        res=0
        for p in self._peers:
            rec=self.allReceptionsInPeer(p.getName())
            restmp=self.checkAllEmissionsToPeer(p.getName(),rec)
            if (restmp>res):
                res=restmp
        return res


    # Computes a synchronisation set given a peer and a list of peers
    def computeSynchro(self,p,peers):
        alpha=p.buildAlphabetNoDirection()
        alphas=[]
        for p in peers:
            alphap=p.buildAlphabetNoDirection()
            alphas=self.union(alphas,alphap)
        res=self.intersection(alpha,alphas)
        resEM=[]
        for e in res:
            resEM.append(e+"_EM")
        return resEM

    # Computes the intersection of 2 lists
    def intersection(self,l1,l2):
        l=[]
        for e in l1:
            if e in l2:
                l.append(e)
        return l

    # Generates the SVL synchronous composition
    def generateComposition(self, f, synchronousComposition=False, hide=True, bound=1):
        max=len(self._peers)
        i=0
        for p in self._peers:
            if (i+1)<max:
                f.write("par ")
                # synchro set computation
                synchro=self.computeSynchro(p,self._peers[i+1:])
                if synchro!=[]:
                    self.dumpStringList(f,synchro)
                    f.write(" in\n")

            if not(synchronousComposition):
                alphap=p.buildAlphabet()
                alphaqueue=p.buildAlphabetQueue()
                alphaqueuebis=self.addBothDirections(alphaqueue)
                alphap=self.union(alphap,alphaqueuebis)
                if hide:
                    f.write("( total hide ")
                    p.dumpAlphabetWithREC(alphaqueue, f, False)
                    f.write(" in ")
                f.write("peer_queue_")
            else:
                alphap=p.buildAlphabetEmissionOnly()

            f.write(p.getName())
            if not(synchronousComposition):
                f.write("_"+str(bound))
            p.dumpAlphabet(alphap,f,False)
            if not(synchronousComposition) and hide:
                f.write(") ")
            
            i=i+1
            if (i<max):
                f.write("\n||\n")
        i=1
        f.write("\n")
        while (i<max):
           f.write("end par\n")
           i=i+1

        f.write (";\n\n")


    # Generates an SVL file
    def generateSVL(self, smartComposition=True, synchronizability=True, bound=0, increment=1, equiv="branching", pmodel="send"):
        name=self._name
        filename=mytmp+name+".svl"
        f=open(filename, 'w')
        hide=(pmodel=="send")
        f.write("% CAESAR_OPEN_OPTIONS=\"-silent -warning\"\n% CAESAR_OPTIONS=\"-more cat\"% CADP_TIME=\"/usr/bin/time\"\n\n")
        f.write ("% DEFAULT_PROCESS_FILE="+name+".lnt\n\n")

        if synchronizability:
            if not smartComposition:
                f.write("\""+name+"_compo_sync.bcg\" = root leaf "+equiv+" reduction of\n")
            else:
                f.write("\""+name+"_compo_sync.bcg\" = smart "+equiv+" reduction of\n")
            self.generateComposition(f, True, True) 

            f.write("\""+name+"_compo_sync_min.bcg\"= \""+name+"_compo_sync.bcg\" ;\n\n")

            if not smartComposition:
                f.write("\""+name+"_compo_async.bcg\" = root leaf "+equiv+" reduction of\n")
            else:
                f.write("\""+name+"_compo_async.bcg\" = smart "+equiv+" reduction of\n")
            self.generateComposition(f, False, hide, 1)

            f.write("\""+name+"_compo_async_min.bcg\"= \""+name+"_compo_async.bcg\" ;\n\n")

            # branching synchronizability check 
            f.write("\"synchronizability.bcg\" = "+equiv+" comparison using bfs with bisimulator \""+name+"_compo_sync_min.bcg\" == \""+name+"_compo_async_min.bcg\"");

        else:
            if not smartComposition:
                f.write("\""+name+"_compo_async_"+str(bound)+".bcg\" = root leaf "+equiv+" reduction of\n")
            else:
                f.write("\""+name+"_compo_async_"+str(bound)+".bcg\" = smart "+equiv+" reduction of\n")
            self.generateComposition(f, False, hide, bound)

            f.write("\""+name+"_compo_async_"+str(bound)+"_min.bcg\"= \""+name+"_compo_async_"+str(bound)+".bcg\" ;\n\n")

            if not smartComposition:
                f.write("\""+name+"_compo_async_"+str(bound+increment)+".bcg\" = root leaf "+equiv+" reduction of\n")
            else:
                f.write("\""+name+"_compo_async_"+str(bound+increment)+".bcg\" = smart "+equiv+" reduction of\n")
            self.generateComposition(f, False, hide, bound+increment)

            f.write("\""+name+"_compo_async_"+str(bound+increment)+"_min.bcg\"= \""+name+"_compo_async_"+str(bound+increment)+".bcg\" ;\n\n")

            # branching equivalence check 
            f.write("\"stability.bcg\" = "+equiv+" comparison using bfs with bisimulator \""+name+"_compo_async_"+str(bound)+"_min.bcg\" == \""+name+"_compo_async_"+str(bound+increment)+"_min.bcg\"");
            
        f.close()

    # Generates an SVL file for computing the 1-bounded asynchronous compo (brute force check) 
    def generateSVLOneBounded(self, smartComposition=True, equiv="branching", pmodel="send"):
        name=self._name
        filename=mytmp+name+".svl"
        f=open(filename, 'w')
        hide=(pmodel=="send")
        f.write("% CAESAR_OPEN_OPTIONS=\"-silent -warning\"\n% CAESAR_OPTIONS=\"-more cat\"% CADP_TIME=\"/usr/bin/time\"\n\n")
        f.write ("% DEFAULT_PROCESS_FILE="+name+".lnt\n\n")

        if not smartComposition:
            f.write("\""+name+"_compo_async.bcg\" = root leaf "+equiv+" reduction of\n")
        else:
            f.write("\""+name+"_compo_async.bcg\" = smart "+equiv+" reduction of\n")
        self.generateComposition(f, False, hide, 1)

        f.write("\""+name+"_compo_async_1_min.bcg\"= \""+name+"_compo_async.bcg\" ;\n\n")
        f.close()


    # Generates an SVL file for the brute force check
    def generateSVLBruteForce(self, smartComposition=True, bound=1, increment=1, equiv="branching", pmodel="send"):
        name=self._name
        filename=mytmp+name+".svl"
        f=open(filename, 'w')
        hide=(pmodel=="send")
        f.write("% CAESAR_OPEN_OPTIONS=\"-silent -warning\"\n% CAESAR_OPTIONS=\"-more cat\"% CADP_TIME=\"/usr/bin/time\"\n\n")
        f.write ("% DEFAULT_PROCESS_FILE="+name+".lnt\n\n")

        if not smartComposition:
            f.write("\""+name+"_compo_async_"+str(bound+increment)+".bcg\" = root leaf "+equiv+" reduction of\n")
        else:
            f.write("\""+name+"_compo_async_"+str(bound+increment)+".bcg\" = smart "+equiv+" reduction of\n")
        self.generateComposition(f, False, hide, bound+increment)

        f.write("\""+name+"_compo_async_"+str(bound+increment)+"_min.bcg\"= \""+name+"_compo_async_"+str(bound+increment)+".bcg\" ;\n\n")

        asyncname=name+"_compo_async_"+str(bound)+"_min"
        f.write("\"stability.bcg\" = "+equiv+" comparison using bfs with bisimulator \""+asyncname+".bcg\" == \""+name+"_compo_async_"+str(bound+increment)+"_min.bcg\"");
            
        f.close()

    # Generates an SVL file for the brute force check (decrementing by increment)
    def generateSVLBruteForceDec(self, smartComposition=True, bound=1, increment=1, equiv="branching", pmodel="send"):
        name=self._name
        filename=mytmp+name+".svl"
        f=open(filename, 'w')
        hide=(pmodel=="send")
        f.write("% CAESAR_OPEN_OPTIONS=\"-silent -warning\"\n% CAESAR_OPTIONS=\"-more cat\"% CADP_TIME=\"/usr/bin/time\"\n\n")
        f.write ("% DEFAULT_PROCESS_FILE="+name+".lnt\n\n")

        if not smartComposition:
            f.write("\""+name+"_compo_async_"+str(bound)+".bcg\" = root leaf "+equiv+" reduction of\n")
        else:
            f.write("\""+name+"_compo_async_"+str(bound)+".bcg\" = smart "+equiv+" reduction of\n")
        self.generateComposition(f, False, hide, bound)

        f.write("\""+name+"_compo_async_"+str(bound)+"_min.bcg\"= \""+name+"_compo_async_"+str(bound)+".bcg\" ;\n\n")

        asyncname=name+"_compo_async_"+str(bound)+"_min"
        f.write("\"stability.bcg\" =  "+equiv+" comparison using bfs with bisimulator \""+asyncname+".bcg\" == \""+name+"_compo_async_"+str(bound+increment)+"_min.bcg\"");
            
        f.close()


    # gets the number of peer in that directory
    def getNbPeers(self):
        call("ls "+self._name+"/peer*.aut | wc -l >& res.txt", shell=True)
        nbpeers=open("res.txt",'r').readline().strip(' ').strip('\n')
        call("rm res.txt", shell=True)
        return nbpeers



    # whole methodology: synchronizability -> k:=1 -> stability check
    # if false then increments one by one until reaching kmax
    def checkBruteForce(self, nbpeers, increment=1, kmax=10, equivnotion="branching", cmodel="fifo", pmodel="send"):
        # transformation from aut files to python model for peers
        async.aut2py4peers(nbpeers)
        if (pmodel=="send"):
            # composition generation and synchronizability check
            async.peers2lnt(nbpeers,self._name,1,True,increment,cmodel) # bound to 1 for synchronizability
            async.generateSVL(False, True, equiv=equivnotion, pmodel=pmodel)
            call('cd '+mytmp+' ; svl '+self._name+'.svl', shell=True)
            res=call('grep TRUE '+mytmp+self._name+'.log', shell=True)
        else:
            # we do not check synchronizability for both send/receive msgs in models
            res=1 
            # but we generate the 1-bounded asynchronous system
            async.peers2lnt(nbpeers,self._name,1,True,increment,cmodel) # bound to 1 for synchronizability
            async.generateSVLOneBounded(False, equiv=equivnotion, pmodel=pmodel)
            call('cd '+mytmp+' ; svl '+self._name+'.svl', shell=True)
            res=call('grep TRUE '+mytmp+self._name+'.log', shell=True)

        if (res==1):
            if (pmodel=="send"):
                print "This system is not synchronizable\n"

            k=1
            cyc=0
            if (cyc!=-1):
                stopcond=1
                while (stopcond==1) and (k<=kmax):
                    # print "k: "+str(k)
                    async.peers2lnt(nbpeers,self._name,k,False,increment,cmodel) # bound to k and k+increment
                    async.generateSVLBruteForce(False, k, increment,equivnotion,pmodel)
                    call('cd '+mytmp+' ; svl '+self._name+'.svl', shell=True)
                    stopcond=call('grep TRUE '+mytmp+self._name+'.log', shell=True)
                    k=k+1
                if (k==(kmax+1)):
                    print "Inconclusive response, equivalence false for kmax "+str(kmax)
                else:
                    k=k-1
                    print "System stabilizing at bound "+str(k)
            else:
                k=-1
                print "Verification is not possible, there is a cycle of emissions in one of the peers\n"
        else:
            k=1
            print "This system is synchronizable\n"
        infotxt=open("info.txt", 'w')
        infotxt.write(str(k)+"\n")
        infotxt.close()
        cdmv="mv info.txt "+self._name
	call(cdmv, shell=True)

    # whole methodology: synchronizability -> k computation -> stability check
    # depending of the result increments or decrements one by one
    def checkBruteForceFromK(self, nbpeers, increment=1, kmax=10, equivnotion="branching", cmodel="fifo", pmodel="send", initk="k2"):
        # transformation from aut files to python model for peers
        async.aut2py4peers(nbpeers)
        if (pmodel=="send"):
            # composition generation and synchronizability check
            async.peers2lnt(nbpeers,self._name,1,True,increment,cmodel) # bound to 1 for synchronizability
            async.generateSVL(False, True, equiv=equivnotion, pmodel=pmodel)
            call('cd '+mytmp+' ; svl '+self._name+'.svl', shell=True)
            res=call('grep TRUE '+mytmp+self._name+'.log', shell=True)
        else:
            # this does not make sense to check synchronizability if we keep send and receive msgs in models !
            res=1
        if (res==1):
            if (pmodel=="send"):
                print "This system is not synchronizable\n"

            cyc=0
            if (cyc!=-1):

                if (initk=="k1"):
                    k=async.longestEmissionPathPeers() # replace this computation depending of the experiments
                else:
                    k1=async.longestEmissionPathPeers() # max longest em seq or received emis.
                    k2=async.numberOfReceivedEmissionsPeers()
                    if (k1>k2):
                        k=k1
                    else:
                        k=k2

                print "k: "+str(k)
                async.peers2lnt(nbpeers,self._name,k,False,increment,cmodel) # bound to k and k+increment
                async.generateSVL(False, False, k, increment, equivnotion, pmodel)
                call('cd '+mytmp+' ; svl '+self._name+'.svl', shell=True)
                stopcond=call('grep TRUE '+mytmp+self._name+'.log', shell=True)
                if (stopcond==1): # we are looking for a larger bound
                    k=k+1
                    while (stopcond==1) and (k<=kmax):
                        print "k: "+str(k)
                        async.peers2lnt(nbpeers,self._name,k,False,increment,cmodel) # bound to k and k+increment
                        async.generateSVLBruteForce(False, k, increment, equivnotion, pmodel)
                        call('cd '+mytmp+' ; svl '+self._name+'.svl', shell=True)
                        stopcond=call('grep TRUE '+mytmp+self._name+'.log', shell=True)
                        k=k+1
                    k=k-1
                else: # we are looking for smaller bound
                    if (k>1):
                        k=k-1
                        while (stopcond==0) and (k>0): 
                            print "k: "+str(k)
                            async.peers2lnt(nbpeers,self._name,k,False,increment,cmodel) # bound to k and k+increment
                            async.generateSVLBruteForceDec(False, k, increment, equivnotion, pmodel)
                            call('cd '+mytmp+' ; svl '+self._name+'.svl', shell=True)
                            stopcond=call('grep TRUE '+mytmp+self._name+'.log', shell=True)
                            k=k-1
                        k=k+2 
                if (k==kmax):
                    print "Inconclusive response, equivalence false for kmax "+str(kmax)
                else:
                    print "System stabilizing at bound "+str(k)
            else:
                k=-1
                print "Verification is not possible, there is a cycle of emissions in one of the peers\n"
        else:
            k=1
            print "This system is synchronizable\n"
        infotxt=open("info.txt", 'w')
        infotxt.write(str(k)+"\n")
        infotxt.close()
        cdmv="mv info.txt "+self._name
	call(cdmv, shell=True)


    # whole methodology: synchronizability -> k computation -> stability check
    # the first 'k' is computed heuristically + binary search
    def checkBinarySearch(self, nbpeers, increment=1, kmax=10, equivnotion="branching", cmodel="fifo", pmodel="send", initk="k2"):
        # transformation from aut files to python model for peers
        async.aut2py4peers(nbpeers)
        if (pmodel=="send"):
            # composition generation and synchronizability check
            async.peers2lnt(nbpeers,self._name,1,True,increment,cmodel) # bound to 1 for synchronizability
            async.generateSVL(True, True, equiv=equivnotion, pmodel=pmodel)
            call('cd '+mytmp+' ; svl '+self._name+'.svl', shell=True)
            res=call('grep TRUE '+mytmp+self._name+'.log', shell=True)
        else:
            res=1 # we do not check synchronizability for both send/receive msgs in models
        if (res==1):
            if (pmodel=="send"):
                print "This system is not synchronizable\n"
            cyc=0
            if (cyc!=-1):

                if (initk=="k1"):
                    k=async.longestEmissionPathPeers()
                else:
                    k1=async.longestEmissionPathPeers() # max longest em seq or received emis.
                    k2=async.numberOfReceivedEmissionsPeers()
                    if (k1>k2):
                        k=k1
                    else:
                        k=k2

                sol=kmax

                lowerk=1
                upperk=kmax
                stopcond=1
                undecided=False

                print "lowerk: "+str(lowerk)
                print "upperk: "+str(upperk)
                print "k: "+str(k)
                
                while (stopcond==1) and not(undecided) and (k!=sol) and (k!=0) and (lowerk<=upperk):
                    
                    async.peers2lnt(nbpeers,self._name,k,False,increment,cmodel) # bound to k and k+increment
                    async.generateSVL(True, False, k, increment, equivnotion, pmodel)

                    call('cd '+mytmp+' ; svl '+self._name+'.svl', shell=True)
                    stopcond=call('grep TRUE '+mytmp+self._name+'.log', shell=True)

                    # dichotomic binary search
                    if (stopcond!=1): # equivalence is true, bound is between lowerk and k
                        if (k<sol):
                            sol=k
                        upperk=k-1 
                        ktmp=k
                        k=((lowerk+upperk)/2)
                        if (k!=ktmp): # we stop when we keep on checking (positively) the same bound
                            stopcond=1
                    else: # equivalence is false, bound is between k and upperk
                        lowerk=k+1 
                        k=((lowerk+upperk)/2)+1
                        if (k==kmax): # undecidable for kmax
                            undecided=True

                    print "lowerk: "+str(lowerk)
                    print "upperk: "+str(upperk)
                    print "k: "+str(k)

                if undecided:
                    print "Inconclusive response, equivalence false for kmax "+str(kmax)
                else:
                    print "System stabilizing at bound "+str(sol)
            else:
                k=-1
                print "Verification is not possible, there is a cycle of emissions in one of the peers\n"
        else:
            k=1
            print "This system is synchronizable\n"
        infotxt=open("info.txt", 'w')
        infotxt.write(str(k)+"\n")
        infotxt.close()
        cdmv="mv info.txt "+self._name
	call(cdmv, shell=True)

    # compares own bound computation with experimental bound (info.txt)
    def compareBufferBounds(self, nbpeers):
        # transformation from aut files to python model for peers
        async.aut2py4peers(nbpeers)

        k=async.numberOfReceivedEmissionsPeers()

        # alternative k computations

        #k1=async.numberOfReceivedEmissionsPeers() 
        #k2=async.longestEmissionPathPeers()
        #if (k1>k2):
        #    k=k1
        #else:
        #    k=k2
        #k=async.numberOfEmissionsPeers()


        finfo=open(self._name+"/info.txt", 'r')
        kexp=finfo.read()
        finfo.close()
        if (kexp.strip()=="0"):
            print self._name+": synchronizable"
        elif (kexp.strip()=="-1"):
            print self._name+": cycle of emissions"
        else:
            res=self._name+": own computation bound "+str(k)+" vs experimental bound "+kexp.strip()
            if (k>int(kexp.strip())) and (int(kexp.strip())!=10):
                res=res+" -> PROBLEM HERE!"
            print res

    # Checks equivalence for the asynchronous system for bounds from k to k+incmax
    #def checkStability(self, nbpeers, increment=1, incmax=3):
        finfo=open(self._name+"/info.txt", 'r')
        kexp=finfo.read()
        finfo.close()
        k=int(kexp.strip())
        print self._name+": "+str(k)
        if ((k!=-1) and (k!=0) and (k!=10)):
            bmax=k+incmax
            stopcond=0
            async.aut2py4peers(nbpeers)
            while (stopcond==0) and (k<bmax): 
                async.peers2lnt(nbpeers,self._name,k,False,increment) # bound and bound+increment
                async.generateSVL(True, False, k, increment)
                call('cd '+mytmp+' ; svl '+self._name+'.svl', shell=True)
                stopcond=call('grep TRUE '+mytmp+self._name+'.log', shell=True)
                k=k+increment

    # seeks a peer containing a message with a specific direction
    def getPeer(self,message,direction):
        for p in self._peers:
            for t in p._T:
               if (t.getLabel().getMessage()==message) and (t.getLabel().getDirection()==direction):
                   return p._name

    # Dumps from the current example/system a txt file respecting ASE'14 tool support input format
    def dumpTxtFile(self, nbpeers): 
        # transformation from aut files to python model for peers
        async.aut2py4peers(nbpeers)
        filename=mytmp+self._name+".txt"
        f=open(filename, 'w')
        f.write(" %% example "+self._name+" \n\n")
        for p in self._peers:
            for t in p._T:
                f.write("ptrans("+p._name+", "+p._name+t.getSource().getName()+", ")
                if (t.getLabel().getMessage()=="i"):
                    f.write("(epsilon), ")
                else:
                    if (t.getLabel().getDirection()=="EM"):
                        f.write("out(")
                    else:
                        f.write("in(")
                    partner=self.getPeer(t.getLabel().getMessage(),t.getLabel().opposite())
                    f.write(t.getLabel().getMessage()+","+partner+"), ")
                f.write(p._name+t.getTarget().getName()+").\n")
            f.write("startPeer("+p._name+","+p._name+p._s0.getName()+").\n\n")
        f.close()

# This methods checks whether the parameters of the stabc script are valid or not
def check_param(param):

    # check if the number of parameter is correct
    if (len(param)<7):
        print "\nError: not enough parameters.\n"
        return False
    if (len(param)>7):
        print "\nError: too many parameters.\n"
        return False


    # check whether the directory name exists
    directory=param[1]
    from subprocess import call
    tmp=call(["ls", param[1]])
    if (tmp==1):
        print "\nError: the directory \""+param[1]+"\" does not exist.\n"
        return False

    # check whether the equivalence notion used is valid
    tmp=param[2]
    if not((tmp=="strong") or (tmp=="branching") or (tmp=="trace")):
        print "\nError: the equivalence notion \""+tmp+"\" is not supported.\n"
        return False

    # check whether the communication model used is valid
    tmp=param[3]
    if not((tmp=="fifo") or (tmp=="bag")):
        print "\nError: the communication model \""+tmp+"\" is not supported.\n"
        return False

    # check whether the peer / composition model used is valid
    tmp=param[4]
    if not((tmp=="send") or (tmp=="all")):
        print "\nError: the peer model \""+tmp+"\" is not valid.\n"
        return False

    # check whether the strategy used is valid
    tmp=param[5]
    if ((int(tmp)<1) or (int(tmp)>5)):
        print "\nError: the strategy code \""+tmp+"\" is not valid.\n"
        return False

    # if all parameters are correct, returns true
    return True


if __name__ == '__main__':

    import sys
    from subprocess import call

    increment=1

    if check_param(sys.argv):
        async=Async(sys.argv[1])
        nbpeers=async.getNbPeers()
        equiv=sys.argv[2]          # equivalence notion used for comparison (strong, branching, trace)
        cmodel=sys.argv[3]         # communication model (fifo, bag)
        pmodel=sys.argv[4]         # peer model (send, all)
        strategy=int(sys.argv[5])  # strategy (1, 2, 3, 4, 5)
        kmax=int(sys.argv[6])      # kmax (arbitrary bound to stop the stability check)

        if (strategy==1):
            async.checkBruteForce(int(nbpeers),increment,kmax,equiv,cmodel,pmodel)
        elif (strategy==2):
            async.checkBinarySearch(int(nbpeers),increment,kmax,equiv,cmodel,pmodel,"k1")
        elif (strategy==4):
            async.checkBinarySearch(int(nbpeers),increment,kmax,equiv,cmodel,pmodel,"k2")
        elif (strategy==3):
            async.checkBruteForceFromK(int(nbpeers),increment,kmax,equiv,cmodel,pmodel,"k1")
        else: # strategy is 5
            async.checkBruteForceFromK(int(nbpeers),increment,kmax,equiv,cmodel,pmodel,"k2")

    else:
        print "Please use the following command: python stabc.py DirectoryName EquivNotion CommModel PeerModel Strategy KMax \n"
        print "   where EquivNotion = strong | branching | trace"
        print "         CommModel   = fifo | bag"
        print "         PeerModel   = send | all"
        print "         Strategy    = 1 | 2 | 3 | 4 | 5"
        print "         KMax        = any integer \n"
        print "Example: python stabc.py EX150 branching fifo all 1 10 (see README for details)\n"

